--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.STD_LOGIC_1164.all;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Package Definition
Package fir_utils Is

   -- constant declaration
   Constant NUM_TAPS : integer := .Configurable_ReplaceWith_NumTaps;

   Type coeff_inputs Is Array (0 To NUM_TAPS-1) of std_logic_vector(.Configurable_ReplaceWith_WidthCo-1 downto 0);

   -- type definition
   Type x_array Is Array (0 to NUM_TAPS)   of std_logic_vector(.Configurable_ReplaceWith_WidthX-1 downto 0);
   Type s_array Is Array (0 to NUM_TAPS)   of std_logic_vector(.Configurable_ReplaceWith_WidthY-1 downto 0);
   Type h_array Is Array (0 to NUM_TAPS-1) of std_logic_vector(.Configurable_ReplaceWith_WidthCo-1 downto 0);

   -- component declarations
   Component tap
       Generic (
           WIDTH_X : Natural;
           WIDTH_Y : Natural;
           WIDTH_H : Natural
       );
       Port (
           clk      : in  std_logic;
           ce       : in  std_logic;
           xn       : in  std_logic_vector(WIDTH_X-1 downto 0);
           xn_out   : out std_logic_vector(WIDTH_X-1 downto 0);
           hn       : in  std_logic_vector(WIDTH_H-1 downto 0);
           prev_sum : in  std_logic_vector(WIDTH_Y-1 downto 0);
           sum_out  : out std_logic_vector(WIDTH_Y-1 downto 0)
       );
   End Component;

end package fir_utils;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
package body fir_utils is
end package body fir_utils;
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.STD_LOGIC_1164.all;
Use IEEE.STD_LOGIC_ARITH.all;
Use IEEE.STD_LOGIC_SIGNED.all;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
Entity tap Is
    Generic (
        WIDTH_X : Natural;
        WIDTH_Y : Natural;
        WIDTH_H : Natural
    );
    Port (
        clk      : in  std_logic;
        ce       : in  std_logic;
        xn       : in  std_logic_vector(WIDTH_X-1 downto 0);
        xn_out   : out std_logic_vector(WIDTH_X-1 downto 0);
        hn       : in  std_logic_vector(WIDTH_H-1 downto 0);
        prev_sum : in  std_logic_vector(WIDTH_Y-1 downto 0);
        sum_out  : out std_logic_vector(WIDTH_Y-1 downto 0)
    );
end entity tap;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
Architecture behav Of tap Is
    Signal product  : std_logic_vector((WIDTH_X+WIDTH_H)-1 downto 0);
    Signal xn_delay : std_logic_vector(WIDTH_X-1 downto 0);
Begin

    reg : Process (clk) Is
    Begin
       If (clk'event AND clk = '1') Then
          If (ce = '1') Then
              xn_delay <= xn;
          End If;
       End If;         
    End Process;

    xn_out  <= xn_delay;
    product <= signed(xn_delay) * signed(hn);
    sum_out <= signed(prev_sum(WIDTH_Y-1 downto 0)) + signed(product(WIDTH_Y-1 downto 0));

End behav;
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.STD_LOGIC_1164.all;

Library work;
Use work.fir_utils.all;

--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
Entity .Configurable_ReplaceWith_Designator Is
    Port(
        clk:        in  std_logic;
        ce:         in  std_logic;
        Xn:         in  std_logic_vector(.Configurable_ReplaceWith_WidthX - 1 downto 0);
        Yn:         out std_logic_vector(.Configurable_ReplaceWith_WidthY - 1 downto 0);
        .Configurable_ReplaceWith_CoeffPorts
    );
End .Configurable_ReplaceWith_Designator;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
Architecture struct Of .Configurable_ReplaceWith_Designator Is
    Signal x_sig    : x_array;
    Signal sum_sig  : s_array;
    Signal h_sig    : h_array;
Begin

    x_sig(0) <= Xn;
    sum_sig(0) <= (OTHERS=>'0');
    .Configurable_ReplaceWith_CoeffAssignments

    taps: For iTap In 0 To NUM_TAPS-1 Generate
        atap : tap
        Generic Map (WIDTH_X=>.Configurable_ReplaceWith_WidthX, WIDTH_Y=>.Configurable_ReplaceWith_WidthY, WIDTH_H=>.Configurable_ReplaceWith_WidthCo)
        Port Map(
            clk      => clk,
            ce       => ce,
            xn       => x_sig(iTap),
            xn_out   => x_sig(iTap+1),
            hn       => h_sig(iTap),
            prev_sum => sum_sig(iTap),
            sum_out  => sum_sig(iTap+1));
    End Generate;

    output_reg : Process (clk) Is
    Begin
       If (clk'event AND clk = '1') Then
        Yn <= sum_sig(NUM_TAPS);
       End If;
    End Process;
End struct;
--------------------------------------------------------------------------------


